import os
import numpy as np
import gdown
from scipy.special import erf
from functools import wraps
from scipy import interpolate
from scipy.ndimage import median_filter,generic_filter

def datadir():
    """ Return the data/ directory."""
    fil = os.path.abspath(__file__)
    codedir = os.path.dirname(fil)
    datadir = os.path.abspath(codedir+'../../../data/')+'/'
    return datadir

def nanmedfilt(x,size,mode='reflect'):
    return generic_filter(x, np.nanmedian, size=size)

def gauss2dbin(x,amp,center,sigma):
    """ Make lots of Gaussian profiles."""
    # amp, center, sigma are arrays
    # x should have shape [Npix, Ngaussians]
    
    xcen = x-center.reshape(1,-1)
    dx = 1
    x1cen = xcen - 0.5*dx  # left side of bin
    x2cen = xcen + 0.5*dx  # right side of bin

    t1cen = x1cen/(np.sqrt(2.0)*sigma.reshape(1,-1))  # scale to a unitless Gaussian
    t2cen = x2cen/(np.sqrt(2.0)*sigma.reshape(1,-1))

    # For each value we need to calculate two integrals
    #  one on the left side and one on the right side

    # Evaluate each point
    #   ERF = 2/sqrt(pi) * Integral(t=0-z) exp(-t^2) dt
    #   negative for negative z
    geval_lower = erf(t1cen)
    geval_upper = erf(t2cen)

    geval = (amp*sigma).reshape(1,-1) * np.sqrt(2.0) * np.sqrt(np.pi)/2.0 * ( geval_upper - geval_lower )

    return geval

def weightedregression(x,y,w,axis=0,zero=False):
    """ Perform weighted simple linear regression on 2D data."""
    # zero means the intercept = 0, y = m*x
    
    if x.ndim==1:
        axis = None
    n = np.nansum(np.isfinite(x)*np.isfinite(y),axis=axis)

    # Relative weights
    #  only count x/y points that are not NaN
    totwt = np.nansum(w*np.isfinite(x)*np.isfinite(y),axis=axis)
    wt = w/totwt
            
    # Solve for intercept as well
    if zero==False:
        # Simple Linear Regression
        # https://en.wikipedia.org/wiki/Simple_linear_regression
        # y = m*x+b
        # m = (N*Sum(x*y)-Sum(x)*Sum(y))/(N*Sum(x^2)-Sum(x)^2)
        
        # Weighted version
        # https://stats.stackexchange.com/questions/12673/formula-for-weighted-simple-linear-regression
        # xwtmn = Sum(w*x)/Sum(w)
        # ywtmn = Sum(w*y)/Sum(w)
        # m = Sum(w*(x-xwtmn)*(y-ywtmn)) / Sum(w*(x-xwtmn)^2)
        # b = ywtmn - m*xwtmn

        # https://ms.mcmaster.ca/canty/teaching/stat3a03/Lectures7.pdf
        
        xwtmn = np.nansum(wt*x,axis=axis)
        ywtmn = np.nansum(wt*y,axis=axis)
        m = np.nansum(wt*(x-xwtmn)*(y-ywtmn),axis=axis) / np.nansum(wt*(x-xwtmn)**2,axis=axis)
        b = ywtmn - m*xwtmn

        # Uncertainties
        # var(m) = sig^2 / Sum(w*(x-xwtmn)^2)
        # var(b) = sig^2 * (1/Sum(w) + xwtmn^2/Sum(w*(x-xwtmn)^2))
        sig = np.sqrt( np.nansum(wt*(y-b-m*x)**2,axis=axis)/(n-2) )
        merr = np.sqrt( sig**2 / np.nansum(wt*(x-xwtmn)**2,axis=axis) )
        berr = np.sqrt( sig**2 * (1 + xwtmn**2/np.nansum(wt*(x-xwtmn)**2,axis=axis)) )
        
        return m,merr,b,berr

    # Intercept is zero
    else:
        # simple linear regression without the regression term (single regressor)
        # https://en.wikipedia.org/wiki/Simple_linear_regression
        m = np.nansum(wt*x*y,axis=axis)/np.nansum(wt*x**2,axis=axis)
        sig = np.sqrt( np.nansum(wt*(y-m*x)**2,axis=axis)/(n-1) )
        merr = np.sqrt( sig**2 / np.nansum(wt*x**2,axis=axis) )
        
        return m,merr

def scalarDecorator(func):
    """Decorator to return scalar outputs for wave2pix and pix2wave
    """
    @wraps(func)
    def scalar_wrapper(*args,**kwargs):
        if np.array(args[0]).shape == ():
            scalarOut= True
            newargs= (np.array([args[0]]),)
            for ii in range(1,len(args)):
                newargs= newargs+(args[ii],)
            args= newargs
        else:
            scalarOut= False
        result= func(*args,**kwargs)
        if scalarOut:
            return result[0]
        else:
            return result
    return scalar_wrapper

@scalarDecorator
def wave2pix(wave,wave0):
    """ convert wavelength to pixel given wavelength array
    Args :
       wave(s) : wavelength(s) (\AA) to get pixel of
       wave0 : array with wavelength as a function of pixel number 
    Returns :
       pixel(s) in the chip
    """
    pix0 = np.arange(len(wave0))
    # Need to sort into ascending order
    sindx = np.argsort(wave0)
    wave0 = wave0[sindx]
    pix0 = pix0[sindx]
    # Start from a linear baseline
    #baseline = np.polynomial.Polynomial.fit(wave0,pix0,1)
    #ip = interpolate.InterpolatedUnivariateSpline(wave0,pix0/baseline(wave0),k=3)
    #out = baseline(wave)*ip(wave)
    out = interpolate.InterpolatedUnivariateSpline(wave0,pix0,k=3)(wave) 
    # NaN for out of bounds
    out[wave > wave0[-1]] = np.nan
    out[wave < wave0[0]] = np.nan
    return out

@scalarDecorator
def pix2wave(pix,wave0):
    """ convert pixel(s) to wavelength(s)
    Args :
       pix : pixel(s) to get wavelength at
       wave0 : array with wavelength as a function of pixel number 
    Returns :
       wavelength(s) in \AA
    """
    pix0 = np.arange(len(wave0))
    # Need to sort into ascending order
    sindx = np.argsort(pix0)
    wave0 = wave0[sindx]
    pix0 = pix0[sindx]
    # Start from a linear baseline
    baseline = np.polynomial.Polynomial.fit(pix0,wave0,1)
    ip = interpolate.InterpolatedUnivariateSpline(pix0,wave0/baseline(pix0), k=3)
    out = baseline(pix)*ip(pix)
    # NaN for out of bounds
    out[pix < 0] = np.nan
    out[pix > 2047] = np.nan
    return out


def expand_msa_slits(tab,msa_metadata_id=1,dither_position=1):
    """ Modify the msa shutter table to expand the slits."""

    # A lot of this code was taken from assign_wcs.nirspec.get_open_msa_slits()
    
    # First we are going to filter the msa_file data on the msa_metadata_id                                                                                      
    # and dither_point_index.                                                                                                                                    
    msa_data = [np.array(x) for x in tab if x['msa_metadata_id'] == msa_metadata_id
                and x['dither_point_index'] == dither_position]

    # Get all source_ids for slitlets with sources.                                                                                                              
    # These should not be used when assigning source_id to background slitlets.                                                                                  
    source_ids = set([x[5] for x in tab if x['msa_metadata_id'] == msa_metadata_id
                      and x['dither_point_index'] == dither_position])
    #print(len(source_ids),'sources')

    
    # Get the unique slitlet_ids  
    slitlet_ids_unique = list(set([int(x['slitlet_id']) for x in msa_data]))

    # SDP may assign a value of "-1" to ``slitlet_id`` - these need to be ignored.                                                                               
    # JP-436                                                                                                                                                     
    if -1 in slitlet_ids_unique:
        slitlet_ids_unique.remove(-1)

    newtab = tab.copy()
        
    # Loop over the source_ids
    # Now lets look at each unique slitlet id                                                                                                                    
    for i,slitlet_id in enumerate(slitlet_ids_unique):
        # Get the rows for the current slitlet_id                                                                                                                
        slitlets_sid = [x for x in msa_data if x['slitlet_id'] == slitlet_id]
        open_shutters = [int(x['shutter_column']) for x in slitlets_sid]

        main_shutter = [s for s in slitlets_sid if s['primary_source'] == 'Y']
        n_main_shutter = len([s for s in slitlets_sid if s['primary_source'] == 'Y'])

        xcen, ycen, quadrant, source_xpos, source_ypos = [
                (int(s['shutter_row']), int(s['shutter_column']), int(s['shutter_quadrant']),
                 float(s['estimated_source_in_shutter_x']),
                 float(s['estimated_source_in_shutter_y']))
                for s in slitlets_sid if s['background'] == 'N'][0]
        source_id = int(main_shutter[0]['source_id'])
        
        #print(i+1,xcen,ycen,n_main_shutter,open_shutters)

        # Need three total
        if len(open_shutters):
            #needcols = [ycen-1,ycen,ycen+1]
            needcols = [ycen-2,ycen-1,ycen,ycen+1,ycen+2]
            needcols = [x for x in needcols if x not in open_shutters]
            # Loop over needed shutters
            for x in needcols:
                # slitlet_id, msa_metadata_id, quadrant, shutter_row, shutter_column, source_id, background, shutter_state,
                #   estimated_source_in_shutter_x, estimated_source_in_shutter_y, dither_point_index, primary_source
                newrow = np.array((slitlet_id, msa_metadata_id, quadrant, xcen, x, source_id, 'Y', 'OPEN', np.nan, np.nan, dither_position, 'N'))
                newtab.add_row(newrow)

    # NOTE, this does NOT check to see if there any "conflicts" of the newly added shutters with any existing shutters 
                
    return newtab

def is_binaryfile(filename):
    """ Check if a file is binary."""    
    try:
        with open(filename, 'r', encoding='utf-8') as f:
            dum = f.read()
            return False
    except UnicodeDecodeError: # Found non-text data
        return True  

def download_data(force=False):
    """ Download the data from my Google Drive."""

    # Check if the "done" file is there
    if os.path.exists(datadir()+'done') and force==False:
        return
    
    data = [{'id':'1ROFRWH825gRBJajordcRcLf_oL0f1JEM','output':'sspcoolgrid1.unf'},
            {'id':'13ybYBEy2hV6zi9KWe7Tv3uyboqFsRCG3','output':'sspcoolgrid2.unf'},
            {'id':'1pZx8Vx0J-fb9uFTGVHDk5c-vhe4Qx6SN','output':'sspcoolgrid3.unf'},
            {'id':'1hFYot7T5ZclbmmZDY_M7drE1ED2xKqaI','output':'sspmediumgrid1.unf'},
            {'id':'1B28nje9vDx2yFdNyR5V36d_aQOVKIW9V','output':'sspmediumgrid2.unf'},
            {'id':'1YvBENKuLpWH4Lop6jkPmAb1BanGDJBdm','output':'ssphotgrid1.unf'},
            {'id':'13ZVGiev-CG2RWljE8kjBnvHdc2SbneCv','output':'ssphotgrid2.unf'},
            {'id':'1HCCaK1ET_4H1XRuDe7jRfi7-x1TsK-dJ','output':'ssphotgrid3.unf'},
            {'id':'10iGSwlO-pwz3aisKDSZml3hwGbx634kS','output':'ssphotgrid4.unf'}]
    
    # This should take 2-3 minutes on a good connection
    
    # Do the downloading
    t0 = time.time()
    print('Downloading '+str(len(data))+' PopStarLight data files')
    for i in range(len(data)):
        print(str(i+1)+' '+data[i]['output'])
        fileid = data[i]['id']
        url = f'https://drive.google.com/uc?id={fileid}'
        output = datadir()+data[i]['output']  # save to the data directory
        if os.path.exists(output)==False or force:
            gdown.download(url, output, quiet=False)

    print('All done in {:.1f} seconds'.format(time.time()-t0))




import os
import copy
import numpy as np
import subprocess
import tempfile
import traceback
import shutil
from dlnpyutils import utils as dln
from scipy.optimize import curve_fit
from theborg import emulator
from astropy.table import Table
from doppler.spec1d import Spec1D,continuum
from . import utils

cspeed = 2.99792458e5  # speed of light in km/s

def gridinfo(filename):
    """ Get information about the FERRE grid."""

    headerfile = filename
    
    if os.path.exists(headerfile)==False:
        # check if the hdr file exists
        headerfile = filename.replace('.dat','.hdr')

    if os.path.exists(headerfile)==False:        
        raise FileNotFoundError(headerfile)
    
    # Read the header information
    f = open(headerfile,'r')
    header = []
    line = f.readline().strip()
    while line != '/':
        header.append(line)
        line = f.readline().strip()
    f.close()

    out = {'header':header}

    header = np.char.array(header)
    
    # Parse the information
    for h in header[1:]:
        dum = h.split()
        k = dum[0]
        val = dum[2:]
        if k.startswith('COMMENT'):
            val = ' '.join(val)
        else:
            orig = val.copy()
            val = []
            for v in orig:
                v = v.replace("'","")
                if dln.isnumber(v):
                    v = float(v)
                elif v.isnumeric():
                    v = int(v)
                val.append(v)    
        if len(val)==1:
            val = val[0]
        out[k] = val

    if 'N_OF_DIM' in out.keys():
        out['NDIM'] = int(out['N_OF_DIM'])
    if 'NPIX' in out.keys():
        out['NPIX'] = int(out['NPIX'])
    if 'LOGW' in out.keys():
        out['LOGW'] = int(out['LOGW'])
    if 'VACUUM' in out.keys():
        out['VACUUM'] = int(out['VACUUM'])
    if 'N_P' in out.keys():
        n_p = [int(n) for n in out['N_P']]
        out['N_P'] = n_p
    # Get wave
    npix = out['NPIX']
    w0 = out['WAVE'][0]
    dw = out['WAVE'][1]
    wave = np.arange(npix)*dw+w0
    if out['LOGW']==1:
        wave = 10**wave
    out['WAVELENGTH'] = wave
    # Labels
    labels = []
    for i in range(out['NDIM']):
        label1 = out['LABEL('+str(i+1)+')']
        labels.append(label1)
    out['LABELS'] = labels
    # Array for each label
    for i in range(out['NDIM']):
        name = labels[i]
        llim = out['LLIMITS'][i]
        step = out['STEPS'][i]        
        n_p = out['N_P'][i]
        vals = np.arange(n_p)*step+llim
        out[name] = vals
        
    return out


def interp(pars,wave=None,cont=None,ncont=None,grid='jwstgiant4.dat',
           griddir=None,ferresrc=None):
    """
    Interpolate in the FERRE grid.

    Parameters
    ----------
    pars : numpy array
       Array of parameters/labels.  This can be a 2D array [Nstars,Nlabels].
    wave : numpy array, optional
       Output wavelengths.  By default, all wavelengths of the grid are used.
    cont : int, optional
       FERRE "cont" continuum parameter.
    ncont : int, optional
       FERRE "ncont" continuum parameter.
    grid : str, optional
       The FERRE grid name.
    griddir : str, optional
       Directory of the FERRE grids.
    ferresrc : str, optional
       Filename of FERRE executable.

    Returns
    -------
    out : dict
       Output parameters.  Dictionary with 'wave' and 'flux'.  If a 2D parameter
         is input, then 'flux' will be 2D but 'wave' will still be 1D.

    Example
    -------

    out = interp([5000.0, 2.5, -1.2, 0.2])

    """

    if griddir is None:
        griddir = utils.datadir()
    gridfile = griddir+grid
    if ferresrc is None:
        ferresrc = subprocess.check_output(['which','ferre'])
        if type(ferresrc) is bytes:
            ferresrc = ferresrc.decode().strip()
        if os.path.exists(ferresrc)==False:
            raise Exception('No FERRE executable found')
        #ferresrc = '/Users/nidever/projects/ferre/bin/ferre'
    info = gridinfo(gridfile)
    
    # Number of objects
    if np.array(pars).ndim==2:
        nobj = pars.shape[0]
    else:
        nobj = 1

    # header file
    headerfile = gridfile
    bingrid = False    
    if os.path.exists(headerfile)==False:
        headerfile = gridfile.replace('.dat','.hdr')
        bingrid = True
        
    # Read the header information
    f = open(headerfile,'r')
    header = []
    line = f.readline().strip()
    while line != '/':
        header.append(line)
        line = f.readline().strip()
    f.close()
    
    # Set up temporary directory
    tmpdir = tempfile.mkdtemp(prefix='ferre')
    curdir = os.path.abspath(os.curdir)
    os.chdir(tmpdir)
    
    # Create fitting input file
    gridbase = os.path.basename(gridfile)
    if bingrid:
        os.symlink(gridfile.replace('.dat','.unf'),tmpdir+'/'+gridbase.replace('.dat','.unf'))
        os.symlink(gridfile.replace('.dat','.hdr'),tmpdir+'/'+gridbase.replace('.dat','.hdr'))
    else:
        os.symlink(gridfile,tmpdir+'/'+gridbase)
    lines = []
    lines += ["&LISTA"]
    lines += ["NDIM = 4"]
    lines += ["NOV = 0"]
    lines += ["INDV = 1 2 3 4"]
    lines += ["SYNTHFILE(1) = '"+headerfile+"'"]
    lines += ["F_FORMAT = 1"]
    lines += ["PFILE = 'ferre.ipf'"]
    lines += ["FFILE = 'ferre.frd'"]
    lines += ["OFFILE = 'ferre.mdl'"]    # output best-fit models
    if wave is not None:
        lines += ["WFILE = 'ferre.wav'"]
        lines += ["WINTER = 2"]    # wavelength interpolate the model fluxes
    lines += ["NOBJ = "+str(nobj)]
    if cont is not None:
        lines += ["CONT = "+str(cont)]      # Running mean normalization
        lines += ["NCONT = "+str(ncont)]     # Npixel for running mean
    lines += ["ERRBAR = 1"]
    lines += ["/"]
    dln.writelines('input.nml',lines)

    # Write the .wav file
    if wave is not None:
        wlines = ''.join(['{:14.5E}'.format(w) for w in wave]) 
        dln.writelines('ferre.wav',wlines)
        
    # Write the IPF file
    if np.array(pars).ndim==2:
        flines = []
        for i in range(nobj):
            flines += ['test{:d} {:.3f} {:.3f} {:.3f} {:.3f}'.format(i+1,*pars[i,:])]
    else:
        nstars = 1
        flines = 'test1 {:.3f} {:.3f} {:.3f} {:.3f}'.format(*pars)
    dln.writelines('ferre.ipf',flines)
    
    # Run FERRE
    if os.path.exists('ferre.mdl'): os.remove('ferre.mdl')
    try:
        result = subprocess.check_output([ferresrc],stderr=subprocess.STDOUT)
    except:
        #traceback.print_exc()
        pass
    
    # Read the output
    mlines = dln.readlines('ferre.mdl')
    if nobj==1:
        mlines = mlines[0]
        mflux = np.array(mlines.split()).astype(float)
    else:
        npix = len(np.array(mlines[0].split()).astype(float))
        mflux = np.zeros((nobj,npix),float)
        for i in range(nobj):
            mflux[i,:] = np.array(mlines[i].split()).astype(float)
        
    # Load the wavelengths
    if wave is None:
        wave = info['WAVELENGTH'].copy()

    out = {'wave':wave,'flux':mflux}
    
    # Delete temporary files and directory
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

    return out
    

class FERRE(object):

    def __init__(self,grid='jwstgiant4.dat',outwave=None,cnorder=6,cperclevel=90.0,cbinsize=0.1,
                 loggrelation=False,verbose=False):
        """
        outwave: output wavelength array.  By default, the full grid wavelength array will be used.
        cnorder: continuum normalization polynomial order  Default is 6.
        cperclevel: continuum normalization percentile level for bins.  Default is 90.
        cbinsize: continuum normalization fractional wavelength range (from -1 to 1) to bin.  Default is 0.1.
        """
        gridfile = '/Users/nidever/synspec/winter2017/jwst/'+grid
        ferre = '/Users/nidever/projects/ferre/bin/ferre.x'
        info = gridinfo(gridfile)
        self.grid = grid
        self.gridfile = gridfile
        self.ferre = ferre
        self.info = info
        self.labels = info['LABELS']
        self.nlabels = len(self.labels)
        self.npix = self.info['NPIX']
        self.ranges = np.zeros((self.nlabels,2),float)
        for i in range(self.nlabels):
            self.ranges[i,0] = np.min(self.info[self.labels[i]])
            self.ranges[i,1] = np.max(self.info[self.labels[i]])
        self.steps = self.info['STEPS']
        self.npoints = self.info['N_P']
        self.nspectra = np.sum(self.npoints)
        self.vacuum = self.info['VACUUM']
        self.resolution = self.info['RESOLUTION']
        if outwave is not None:
            self.outwave = outwave
        else:
            self.outwave = None
        self.cnorder = cnorder
        self.cperclevel = cperclevel
        self.cbinsize = cbinsize
        self.loggrelation = loggrelation
        self.verbose = verbose
        self.ncall = 0
        self.njac = 0
        
        # Use logg relation
        if loggrelation:
            # Get logg label
            loggind, = np.where(np.char.array(self.labels).lower()=='logg')
            if len(loggind)==0:                
                raise ValueError('No logg label')
            self.loggind = loggind[0]
            # Get temperature label
            teffind, = np.where(np.char.array(self.labels).lower()=='teff')
            if len(teffind)==0:
                teffind, = np.where(np.char.array(self.labels).lower().find('temp')>-1)
            self.teffind = teffind[0]
            # Get metallicity label
            fehind, = np.where(np.char.array(self.labels).lower()=='feh')
            if len(fehind)==0:
                fehind, = np.where(np.char.array(self.labels).lower()=='metal')
            if len(fehind)==0:
                fehind, = np.where(np.char.array(self.labels).lower()=='mh')                
            if len(fehind)==0:
                fehind, = np.where(np.char.array(self.labels).lower()=='[fe/h]')
            if len(fehind)==0:
                fehind, = np.where(np.char.array(self.labels).lower()=='[m/h]')                
            if len(fehind)==0:
                raise ValueError('No metallicity label')
            self.fehind = fehind[0]
            # Get alpha abundance label
            alphaind, = np.where(np.char.array(self.labels).lower()=='[alpha/fe]')
            if len(alphaind)==0:
                alphaind, = np.where(np.char.array(self.labels).lower()=='alphafe')
            if len(alphaind)==0:
                alphaind, = np.where(np.char.array(self.labels).lower()=='alpha_fe')
            if len(alphaind)==0:
                alphaind, = np.where(np.char.array(self.labels).lower().find('alpha')>-1)
            if len(alphaind)==0:
                self.alphaind = None
            else:
                self.alphaind = alphaind[0]
            # Load the ANN model
            logg_model = emulator.Emulator.load(utils.datadir()+'apogeedr17_rgb_logg_ann.npz')
            self.logg_model = logg_model
            
            
    def __repr__(self):
        """ String representation."""
        out = self.__class__.__name__+'({:s}, {:d} labels, {:d} spectra, {:d} pixels)\n'.format(self.grid,self.nlabels,
                                                                                               self.nspectra,self.npix)
        for i in range(self.nlabels):
            out += '{:8.2f} <= {:^12s} <= {:8.2f}, step {:8.2f} [{:d}]\n'.format(self.ranges[i,0],self.labels[i],
                                                                                 self.ranges[i,1],self.steps[i],self.npoints[i])
        out += '{:8.2f} <= {:^12s} <= {:8.2f}, step {:8.2f} [{:d}]\n'.format(np.min(self.info['WAVELENGTH']),
                                                                             'WAVE',np.max(self.info['WAVELENGTH']),
                                                                             self.info['WAVE'][1],self.npix)
        return out 
        
    def __call__(self,pars,wave=None,cnorder=None,cperclevel=None,cbinsize=None,norm=True):
        """
        pars: array of parameters.  Can be 2D [Nstars,Nlabels].
        """
        if wave is None:
            wave = self.outwave
        if cnorder is None:
            cnorder = self.cnorder
        if cperclevel is None:
            cperclevel = self.cperclevel
        if cbinsize is None:
            cbinsize = self.cbinsize
            
        pars = np.array(pars)
        if pars.ndim==2:
            ninputlabels = pars.shape[1]
        else:
            ninputlabels = len(pars)
        if (ninputlabels != self.nlabels and self.loggrelation==False) or \
           (ninputlabels != (self.nlabels-1) and self.loggrelation==True):
            raise ValueError('Number of labels not correct.')
        
        # Use the logg relation (as a function of Teff/feh/alpha)
        if self.loggrelation:
            # Multiple parameters
            if pars.ndim==2:
                nobj = np.array(pars).shape[0]
                newpars = np.zeros((nobj,self.nlabels),float)
                for i in range(nobj):
                    newpars[i,:] = self.getlogg(pars[i,:])
            # Single set of labels
            else:
                nobj = 1
                newpars = self.getlogg(pars)
        else:
            # Multiple parameters
            if pars.ndim==2:
                nobj = np.array(pars).shape[0]
            else:
                nobj = 1
            newpars = pars

        self.ncall += 1
        out = interp(newpars,wave=wave)

        # Now do the continuum normalization
        if cnorder is not None and norm==True:
            wave = out['wave']
            flux = out['flux'].copy()
            if nobj==1:
                out['flux'] = self.normalize(wave,flux)
            else:
                for i in range(nobj):
                    out['flux'][i,:] = self.normalize(wave,flux[i,:])

        return out
        
    def getlogg(self,pars):
        """ Get logg from the logg relation and fill it into the parameter array where it belongs."""
        # The only parameters should be input, with the logg one missing/excluded
        # Insert a dummy value for logg
        newpars = np.insert(pars,self.loggind,0.0)
        teff = newpars[self.teffind]
        feh = newpars[self.fehind]
        if self.alphaind is not None:
            alpha = newpars[self.alphaind]
        else:
            alpha = 0.0
        logg = self.logg_model([teff,feh,alpha],border='extrapolate')
        newpars[self.loggind] = logg
        return newpars

    def normalize(self,wave,flux,norder=None,perclevel=None,binsize=None):
        """ Normalize a spectrum."""
        if norder is None:
            norder = self.cnorder
        if perclevel is None:
            perclevel = self.cperclevel
        if binsize is None:
            binsize = self.cbinsize
        sp = Spec1D(flux,wave=wave,err=flux*0)
        cont = continuum(sp,norder=norder,perclevel=perclevel,binsize=binsize)
        newflux = flux/cont
        return newflux
    
    def model(self,wave,*pars,**kwargs):
        """ Model function for curve_fit."""
        if self.verbose:
            print('model: ',pars)
        out = self(pars,wave=wave,**kwargs)
        # Only return the flux
        return out['flux']

    def jac(self,wave,*args,retmodel=False,allpars=False,**kwargs):
        """
        Method to return Jacobian matrix.
        This includes the contribution of the lookup table.

        Parameters
        ----------
        wave : numpy array
            Wavelength array.
        args : float
            Model parameter values as separate positional input parameters,
            [amp, xcen, ycen, sky]. If allpars=True, then the model
            parameters are added at the end, i.e. 
            [amp, xcen, ycen, sky, model parameters].
        retmodel : boolean, optional
            Return the model as well.  Default is retmodel=False.

        Returns
        -------
        if retmodel==False
        jac : numpy array
          Jacobian matrix of partial derivatives [N,Npars].
        model : numpy array
          Array of (1-D) model values for the input xdata and parameters.
          If retmodel==True, then (model,jac) are returned.

        Example
        -------

        jac = ferre.jac(xdata,*pars)
        """

        # logg relation
        #  add a dummy logg value in
        if self.loggrelation:
            fullargs = np.insert(args,self.loggind,0.0)
        else:
            fullargs = args

        if self.verbose:
            print('jac: ',args)
            
        # Loop over parameters
        pars = [np.array(copy.deepcopy(args))]
        steps = []
        for i in range(self.nlabels):
            if self.loggrelation and i==self.loggind:
                continue
            targs = np.array(copy.deepcopy(fullargs))
            step = 0.1*self.steps[i]
            # Check boundaries, if above upper boundary
            #   go the opposite way
            if targs[i]>self.ranges[i,1]:
                step *= -1
            targs[i] += step
            # Remove dummy logg if using logg relation
            if self.loggrelation:
                targs = np.delete(targs,self.loggind)
            pars.append(targs)
            steps.append(step)

        # Reformat pars into arrays
        pars = np.array(pars)
        steps = np.array(steps)
            
        # Run FERRE
        out = self(pars,wave=wave)
        flux = out['flux']
        npix = flux.shape[1]
        
        # Initialize jacobian matrix
        fjac = np.zeros((npix,len(steps)),np.float64)
        f0 = flux[0,:]
        for i in range(len(steps)):
            f1 = flux[i+1,:]
            fjac[:,i] = (f1-f0)/steps[i]
            
        self.njac += 1
            
        return fjac

    
def fit(slist,inter=3,algor=1,init=1,indini=None,nruns=1,
        cont=1,ncont=0,errbar=1,grid='jwstgiant4.dat',save=False):
    """ 
    Run FERRE on list of spectra

    Parameters
    ----------
    slist : list
       A list of dictionaries with wave, flux, err and vrel.
    vrel : list
       Input list of doppler shifts for the spectra.
    inter : int, optional
       Interpolation algorithm.
         0   nearest neighbor
         1   linear
         2   quadratic Bezier
         3   cubic Bezier
         4   cubic spline
       Default is 3.
    algor : int, optional
       Search algorithm?
        -1   weighted average over the grid  likely
         0   pixel with lowest value         minlocus
         1   Nelder-mead method (Miller's)   minim
         2   BTR method (Csendes/Miller)     global
         3   uobyqa method (Powell/Miller)    uobyqa
         4   truncated Newton                lmqnbc
         5   MCMC                             mcmcde
       Default is 1.
    init : int, optional
       Starting point for search.
         0   use the values in pfile
         1   follow the rules set by keyword indini
       Default is 1.
    indini : list, optional
       Individual control of starting points for each parameter.
    nruns : int, optional
       Number of searches to be done.  Default is 1.
    cont : int, optional
       Continuum parameter:
         1   polynomial fitting  (ncont order, 0-constant, 1-linear)
         2   segmented normaleization  (ncont segments)
         3   running mean  (ncont pixels)
    ncont : int, optional
       Number for continuum normalization.  If cont=1, then ncont
         gives the polynomial order.  If cont=2, then ncont is the
         number of segements.  If cont=3, then ncont is the number
         of pixels for the running mean.
    errbar : int, optional
       Choice of algorithm to compute error bars.
         0   adopt the distance from solution at which chisq=min(chisq)+1
         1   invert the curvature matrix
       Default is 1.
    save : bool, optional
       Do not delete the temporary directory.  Default is save=False.

    Returns
    -------
    tab : table
       Output table of values.
    info : list
       List of dictionaries, one per spectrum, that includes spectra and arrays.

    Example
    -------

    tab,info = fit(files,vrel,algor=1,nruns=5)

    """
        
    gridfile = '/Users/nidever/synspec/winter2017/jwst/'+grid
    ferre = '/Users/nidever/projects/ferre/bin/ferre.x'

    # Set up temporary directory
    tmpdir = tempfile.mkdtemp(prefix='ferre')
    curdir = os.path.abspath(os.curdir)
    os.chdir(tmpdir)
    print('Running FERRE in temporary directory '+tmpdir)

    # Parameters
    print('INTER = ',inter)
    print('ALGOR = ',algor)
    print('INIT = ',init)
    print('INDINI = ',indini)
    print('NRUNS = ',nruns)
    print('CONT = ',cont)
    print('NCONT = ',ncont)
    print('ERRBAR = ',errbar)
    
    # Create fitting input file
    gridbase = os.path.basename(gridfile)    
    os.symlink(gridfile,tmpdir+'/'+gridbase)
    os.symlink(gridfile.replace('.dat','.unf'),tmpdir+'/'+gridbase.replace('.dat','.unf'))
    os.symlink(gridfile.replace('.dat','.hdr'),tmpdir+'/'+gridbase.replace('.dat','.hdr'))
    lines = []
    lines += ["&LISTA"]
    lines += ["NDIM = 4"]
    lines += ["NOV = 4"]
    lines += ["INDV = 1 2 3 4"]
    lines += ["SYNTHFILE(1) = '"+gridbase+"'"]
    lines += ["F_FORMAT = 1"]
    lines += ["INTER = "+str(inter)]    # cubic Bezier interpolation
    lines += ["ALGOR = "+str(algor)]    # 1-Nelder-Mead,5-MCMC
    lines += ["INIT = "+str(init)]
    if indini is not None:
        lines += ['INDINI = '+np.array2string(np.array(indini)).strip('[]')]        
        nruns = 1
        for term in indini:
            nruns = nruns * term
    #    lines += ["INDINI = "+str(indini)]
    if nruns is not None:
        lines += ["NRUNS = "+str(nruns)]
    lines += ["PFILE = 'ferre.ipf'"]
    lines += ["FFILE = 'ferre.frd'"]
    lines += ["ERFILE = 'ferre.err'"]
    lines += ["WFILE = 'ferre.wav'"]    
    lines += ["OPFILE = 'ferre.opf'"]    # output best-fit parameters and uncertainties
    lines += ["OFFILE = 'ferre.mdl'"]    # output best-fit models
    lines += ["SFFILE = 'ferre.nrd'"]    # normalized data
    lines += ["NOBJ = "+str(len(slist))]
    lines += ["CONT = "+str(cont)]      # Running mean normalization
    lines += ["NCONT = "+str(ncont)]     # Npixel for running mean
    lines += ["WINTER = 2"]    # wavelength interpolate the model fluxes
    lines += ["ERRBAR = "+str(errbar)]
    lines += ["/"]
    dln.writelines('input.nml',lines)
        
    # Put star with maximum npix first
    #  FERRE uses this to set the maximum pixels for ALL the spectra
    for i in range(len(slist)):
        if 'npix' not in slist[i].keys():
            slist[i]['npix'] = len(slist[i]['flux'])
        if 'id' not in slist[i].keys():
            slist[i]['id'] = i+1
    npixall = [s['npix'] for s in slist]
    maxind = np.argmax(npixall)
    npix = np.max(npixall)
    if maxind != 0:
        slist1 = slist.pop(maxind)
        slist = [slist1] + slist

    # Make the list of lines to write out for FERRE
    flines,elines,wlines,plines = [],[],[],[]
    for i in range(len(slist)):
        # Need to buffer the arrays to have consistent number of pixels
        flux = slist[i]['flux']
        err = slist[i]['err']
        wave = slist[i]['wave']        
        if len(flux)<npix:
            nmissing = npix-len(flux)
            flux = np.hstack((flux,np.zeros(nmissing)))
            err = np.hstack((err,np.zeros(nmissing)+1e10))
            wave = np.hstack((wave,np.zeros(nmissing))) 
            
        # Convert to ASCII for FERRE
        obs = ''.join(['{:14.5E}'.format(f) for f in flux])
        obserr = ''.join(['{:14.5E}'.format(e) for e in err])
        obswave = ''.join(['{:14.5E}'.format(w) for w in wave])        
        slist[i]['ferre_fline'] = obs
        slist[i]['ferre_eline'] = obserr
        slist[i]['ferre_wline'] = obswave     
        slist[i]['ferre_pline'] = 'spec'+str(i+1)+' 4000.0  2.5  -0.5  0.1'
        slist[i]['ferre_id'] = 'spec'+str(i+1)
        
        flines.append(slist[i]['ferre_fline'])
        elines.append(slist[i]['ferre_eline'])
        wlines.append(slist[i]['ferre_wline'])
        plines.append(slist[i]['ferre_pline'])
        
    dln.writelines('ferre.frd',flines)
    dln.writelines('ferre.err',elines)
    dln.writelines('ferre.wav',wlines)    
    dln.writelines('ferre.ipf',plines)
    
    # Run FERRE
    if os.path.exists('ferre.mdl'): os.remove('ferre.mdl')
    if os.path.exists('ferre.opf'): os.remove('ferre.opf') 
    print('Running FERRE on '+str(len(slist))+' spectra')
    try:
        fout = open('ferre.log','w')
        out = subprocess.call([ferre],stdout=fout,stderr=subprocess.STDOUT)
        fout.close()
        loglines = dln.readlines('ferre.log')
        #out = subprocess.check_output([ferre])
    except:
        fout.close()        
        #traceback.print_exc()
        pass

    
    # Read the output
    olines = dln.readlines('ferre.opf')
    ids = np.char.array([o.split()[0] for o in olines])
    mlines = dln.readlines('ferre.mdl')
    slines = dln.readlines('ferre.nrd')
    nolines = len(olines)
    print(str(nolines)+' lines in output files')

    # Loop over slist
    for i in range(len(slist)):
        slist1 = slist[i]
        filename = slist1['filename']
        ferre_id = slist1['ferre_id']
        
        # Find the right output line for this star
        ind, = np.where(ids==ferre_id)
        if len(ind)==0:
            print(ferre_id+' not found in output file')
            slist1['success'] = False            
            slist1['pars'] = None
            slist1['parerr'] = None
            slist1['snr2'] = None
            slist1['rchisq'] = None
            slist1['model'] = None
            slist1['smflux'] = None
            continue
        ind = ind[0]
        olines1 = olines[ind]
        mlines1 = mlines[ind]
        slines1 = slines[ind]

        # opfile has name, pars, parerr, fraction of phot data points, log(S/N)^2, log(reduced chisq)
        arr = olines1.split()
        pars = np.array(arr[1:5]).astype(float)
        parerr = np.array(arr[5:9]).astype(float)
        logsnr2 = float(arr[10])
        snr2 = 10**logsnr2
        try:
            logrchisq = float(arr[11])
            rchisq = 10**logrchisq
        except:
            logrchisq = np.nan
            rchisq = np.nan
        slist1['success'] = True            
        slist1['pars'] = pars
        slist1['parerr'] = parerr   
        slist1['snr2'] = snr2
        slist1['rchisq'] = rchisq

        mflux = np.array(mlines1.split()).astype(float)
        smflux = np.array(slines1.split()).astype(float)
        # If this star has less pixels than the first star, then we need to trim some pixels
        #  off the end
        if len(mflux) != slist1['npix']:
            mflux = mflux[0:slist1['npix']]
            smflux = smflux[0:slist1['npix']]
        slist1['model'] = mflux
        slist1['smflux'] = smflux

    # Delete temporary files and directory
    if save==False:    
        os.chdir(curdir)
        shutil.rmtree(tmpdir)

    return slist

def specprep(spec,vrel=None):
    """ Prepare the spectra for input to FERRE """
    # Synspec spectra are in AIR, JWST spectra are in VACUUM, convert
    spec.wavevac = False
    flux = np.array([])
    err = np.array([])
    wave = np.array([])
    mask = np.array([])
    for sp in spec:
        # trim edge pixels
        logood = np.where(sp.mask==False)[0][0]
        higood = np.where(sp.mask==False)[0][-1]
        slc = slice(logood+30,higood-30)
        flux = np.concatenate((flux,sp.flux[slc]))
        err = np.concatenate((err,sp.err[slc]))
        wave = np.concatenate((wave,sp.wave[slc]))
        mask = np.concatenate((mask,sp.mask[slc]))
    # Remove Doppler shift
    if vrel is None:
        if hasattr(spec,'vrel')==False:
            raise ValueError('No vrel to use')
        vrel = spec.vrel
    wave /= (1+vrel/cspeed)
    # Trim wavelengths that are too large
    gd, = np.where(wave < 17990)
    flux = flux[gd]
    err = err[gd]
    wave = wave[gd]
    mask = mask[gd]
    # Grow bad mask by ~20 pixels
    oldmask = mask.copy()
    mask = np.convolve(mask,np.ones(61),mode='same').astype(bool)
    # Interpolate over masked pixels
    xpix = np.arange(len(flux))
    bad, = np.where((mask==True) | (flux<=0))
    good, = np.where((mask==False) & (flux>0))
    if len(bad)>0:
        smfilt = utils.nanmedfilt(flux[good],201)
        newpix = dln.interp(xpix[good],smfilt,xpix[bad],kind='linear')
        flux[bad] = newpix
        #flux[bad] = 0.0
        err[bad] = 1e10

    # Put all the information into slist
    out = {}
    out['npix'] = len(flux)
    out['flux'] = flux
    out['err'] = err
    out['wave'] = wave
    return out


def cfit(slist,vrel,cont=1,ncont=0,loggrelation=False,grid='jwstgiant4.dat',
         initgrid=True,outlier=True,verbose=False):
    """ 
    Fit spectrum with curve_fit running FERRE to get the models
    for each set of parameters.

    Parameters
    ----------
    slist : list
       A list of dictionaries with wave, flux, err and vrel.
    vrel : list
       Input list of doppler shifts for the spectra.
    cont : int, optional
       Continuum parameter:
         1   polynomial fitting  (ncont order, 0-constant, 1-linear)
         2   segmented normaleization  (ncont segments)
         3   running mean  (ncont pixels)
    ncont : int, optional
       Number for continuum normalization.  If cont=1, then ncont
         gives the polynomial order.  If cont=2, then ncont is the
         number of segements.  If cont=3, then ncont is the number
         of pixels for the running mean.
    loggrelation : bool, optional
       Use the logg-relation as a function of Teff/feh/alpha.

    Returns
    -------
    tab : table
       Output table of values.
    info : list
       List of dictionaries, one per spectrum, that includes spectra and arrays.

    Example
    -------

    tab,info = cfit(files,vrel)

    """


    ## Loop over the files
    #slist = []
    #for i in range(nfiles):
    #    filename = files[i]
    #    vrel1 = vrel[i]
    #    if filename is None:
    #        continue
    #    if os.path.exists(filename)==False:
    #        print(filename+' not found')
    #        continue
    #    if os.path.getsize(filename)==0:
    #        print(filename+' is empty')
    #        continue
    #    spec = doppler.read(filename)
    #    spec.vrel = vrel1
    #    #spec.normalize()
    #    slist1 = {'filename':filename,'vrel':vrel1,'snr':spec.snr,'spec':spec}
    #    slist.append(slist1)
    #
    #if len(slist)==0:
    #    print('No spectra to fit')
    #    return None,None

    if type(slist) is not list:
        slist = [slist]
    nspectra = len(slist)
    
    # Loop over spectra
    out = []
    for i in range(nspectra):
        if verbose:
            if i>0: print(' ')
            print('Star '+str(i+1))
            print('-------')
        slist1 = slist[i]
        vrel1 = vrel[i]
        spec = slist1['spec']
        spec.vrel = vrel1
        if verbose:
            print('Vrel: {:.2f} km/s'.format(vrel1))
            print('S/N: {:.2f}'.format(spec.snr))
        # Initialize the FERRE object
        fr = FERRE(loggrelation=loggrelation,verbose=(verbose>1))                
        # Prepare spectrum for FERRE
        pspec = specprep(spec)
        # Now normalize
        newflux = fr.normalize(pspec['wave'],pspec['flux'])
        cont = pspec['flux']/newflux
        pspec['oflux'] = pspec['flux']
        pspec['oerr'] = pspec['err']        
        pspec['flux'] = newflux
        pspec['err'] = pspec['oerr']/cont      
        # Normalize
        fr.outwave = pspec['wave']
        if loggrelation:
            estimates = [4500.0,-1.0,0.0]
            bounds = [np.delete(fr.ranges[:,0],1),np.delete(fr.ranges[:,1],1)]            
        else:
            estimates = [4500.0,2.5,-1.0,0.0]
            bounds = [fr.ranges[:,0],fr.ranges[:,1]]


        # Run grid of ~100 points to get first estimate
        if initgrid:
            if loggrelation:
                nsample = 5
                tstep = np.ptp(fr.ranges[0,:])/nsample
                tgrid = np.arange(nsample)*tstep+fr.ranges[0,0]+tstep*0.5
                mstep = np.ptp(fr.ranges[2,:])/nsample
                mgrid = np.arange(nsample)*mstep+fr.ranges[2,0]+mstep*0.5
                astep = np.ptp(fr.ranges[3,:])/nsample
                agrid = np.arange(nsample)*astep+fr.ranges[3,0]+astep*0.5
                tgrid2d,mgrid2d,agrid2d = np.meshgrid(tgrid,mgrid,agrid)
                gridpars = np.vstack((tgrid2d.flatten(),mgrid2d.flatten(),agrid2d.flatten())).T
            else:
                nsample = 4
                tstep = np.ptp(fr.ranges[0,:])/nsample
                tgrid = np.arange(nsample)*tstep+fr.ranges[0,0]+tstep*0.5
                gstep = np.ptp(fr.ranges[1,:])/nsample
                ggrid = np.arange(nsample)*gstep+fr.ranges[1,0]+gstep*0.5            
                mstep = np.ptp(fr.ranges[2,:])/nsample
                mgrid = np.arange(nsample)*mstep+fr.ranges[2,0]+mstep*0.5
                astep = np.ptp(fr.ranges[3,:])/nsample
                agrid = np.arange(nsample)*astep+fr.ranges[3,0]+astep*0.5  
                tgrid2d,ggrid2d,mgrid2d,agrid2d = np.meshgrid(tgrid,ggrid,mgrid,agrid)
                gridpars = np.vstack((tgrid2d.flatten(),ggrid2d.flatten(),mgrid2d.flatten(),agrid2d.flatten())).T
            if verbose:
                print('Testing an initial grid of '+str(gridpars.shape[0])+' spectra')
            # Run FERRE
            specgrid = fr(gridpars,wave=pspec['wave'])
            chisqarr = np.sum((specgrid['flux']-pspec['flux'])**2/pspec['err']**2,axis=1)/len(pspec['flux'])
            bestind = np.argmin(chisqarr)
            estimates = gridpars[bestind,:]
        if verbose:
            print('Initial estimates: ',estimates)
        
        # Run curve_fit
        try:
            pars,pcov = curve_fit(fr.model,pspec['wave'],pspec['flux'],p0=estimates,
                                  sigma=pspec['err'],bounds=bounds,jac=fr.jac)
            perror = np.sqrt(np.diag(pcov))
            bestmodel = fr.model(pspec['wave'],*pars)
            chisq = np.sum((pspec['flux']-bestmodel)**2/pspec['err']**2)/len(pspec['flux'])

            # Get full parameters
            if loggrelation:
                fullpars = fr.getlogg(pars)
                fullperror = np.insert(perror,fr.loggind,0.0)
            else:
                fullpars = pars
                fullperror = perror

            if verbose:
                printvals = [fullpars[0],fullperror[0],fullpars[1],fullperror[1],fullpars[2],fullperror[2],
                             fullpars[3],fullperror[3]]
                print('Best parameters: {:f}+/-{:.3g}, {:.3f}+/-{:.3g}, {:.3f}+/-{:.3g}, {:.3f}+/-{:.3g}'.format(*printvals))
                print('Chisq: ',chisq)

            # Construct the output dictionary
            out1 = {'index':i,'vrel':vrel1,'snr':spec.snr,'pars':fullpars,'perror':fullperror,'wave':pspec['wave'],
                    'flux':pspec['flux'],'err':pspec['err'],'model':bestmodel,'chisq':chisq,
                    'loggrelation':loggrelation,'success':True}
            success = True
        except:
            traceback.print_exc()
            success = False
            out1 = {'success':False}
                
        # Remove outliers and refit
        if success and outlier:
            diff = pspec['flux']-bestmodel
            med = np.median(diff)
            sig = dln.mad(diff)
            bd, = np.where(np.abs(diff) > 3*sig)
            nbd = len(bd)
            if nbd>0:
                if verbose:
                    print('Removing '+str(nbd)+' outliers and refitting')
                err = pspec['err'].copy()
                flux = pspec['flux'].copy()
                err[bd] = 1e30
                flux[bd] = bestmodel[bd]
                # Save original values
                pars0 = pars
                estimates = pars0
                # Run curve_fit
                try:
                    pars,pcov = curve_fit(fr.model,pspec['wave'],flux,p0=estimates,
                                          sigma=err,bounds=bounds,jac=fr.jac)
                    perror = np.sqrt(np.diag(pcov))
                    bestmodel = fr.model(pspec['wave'],*pars)
                    chisq = np.sum((flux-bestmodel)**2/err**2)/len(flux)

                    # Get full parameters
                    if loggrelation:
                        fullpars = fr.getlogg(pars)
                        fullperror = np.insert(perror,fr.loggind,0.0)
                    else:
                        fullpars = pars
                        fullperror = perror

                    if verbose:
                        printvals = [fullpars[0],fullperror[0],fullpars[1],fullperror[1],fullpars[2],fullperror[2],
                                     fullpars[3],fullperror[3]]
                        print('Best parameters: {:f}+/-{:f}, {:.3f}+/-{:.3f}, {:.3f}+/-{:.3f}, {:.3f}+/-{:.3f}'.format(*printvals))
                        print('Chisq: ',chisq)
                        
                    # Construct the output dictionary
                    out1 = {'index':i,'vrel':vrel1,'snr':spec.snr,'pars':fullpars,'perror':fullperror,'wave':pspec['wave'],
                            'flux':pspec['flux'],'err':pspec['err'],'mflux':flux,'merr':err,'noutlier':nbd,'model':bestmodel,
                            'chisq':chisq,'loggrelation':loggrelation,'success':True}
                    success = True
                except:
                    traceback.print_exc()
                    success = False
                    out1 = {'success':False}
                
        #if verbose and success:
        #    print('Best parameters: ',out1['pars'])
        #    print('Chisq: ',out1['chisq'])
            
        out.append(out1)

    # Create output table
    dt = [('index',int),('vrel',float),('snr',float),('pars',float,4),('perror',float,4),('chisq',float),('success',bool)]
    tab = np.zeros(len(out),dtype=np.dtype(dt))
    for i,o in enumerate(out):
        if o['success']:
            tab['index'][i] = o['index']
            tab['vrel'][i] = o['vrel']
            tab['snr'][i] = o['snr']                        
            tab['pars'][i] = o['pars']	 
            tab['perror'][i] = o['perror']
            tab['chisq'][i] = o['chisq']
            tab['success'][i] = o['success']
    tab = Table(tab)

        
    return out,tab




## Create simple stellar population (SSP) synthetic spectra

import os
import numpy as np
from glob import glob
from dlnpyutils import utils as dln
from astropy.table import Table
from astropy.io import fits
from scipy.interpolate import interp1d
from scipy.stats import binned_statistic_2d
from . import ferre,utils

# Get FERRE grid info
headerfiles = glob(utils.datadir()+'ssp*grid*.hdr')  # Get grid header files
dt = [('filename',str,200),('teffrange',float,2),('loggrange',float,2),
      ('mhrange',float,2),('alpharange',float,2)]
GRIDINFO = np.zeros(len(headerfiles),dtype=np.dtype(dt))
# Get header Teff/logg ranges
for i in range(len(headerfiles)):
    hinfo = ferre.gridinfo(headerfiles[i])
    GRIDINFO['filename'][i] = headerfiles[i]
    GRIDINFO['teffrange'][i] = [np.min(hinfo['TEFF']),np.max(hinfo['TEFF'])]
    GRIDINFO['loggrange'][i] = [np.min(hinfo['LOGG']),np.max(hinfo['LOGG'])]
    GRIDINFO['mhrange'][i] = [np.min(hinfo['[M/H]']),np.max(hinfo['[M/H]'])]
    GRIDINFO['alpharange'][i] = [np.min(hinfo['[ALPHA/M]']),np.max(hinfo['[ALPHA/M]'])]    
# Sort them by temperature
si = np.argsort(GRIDINFO['teffrange'][:,0])
GRIDINFO = GRIDINFO[si]

def ferre_interp(pars):
    """
    Interpolate multiple spectra in grid using FERRE.

    Parameters
    ----------
    pars : numpy array
       Array of parameter values.  Should have dimensions
       of [Nlabels] or [Nstars,Nlabels].  The labels with
       default FERRE synthetic spectral grids should be
       [Teff,logg,[M/H],[alpha/M].

    Returns
    -------
    out : dict
       Dictionary with 1-D wavelength array in "wave", and 2-D
       flux array in "flux" [Nstars,Nwave].

    Example
    -------

    out = ferre_interp(pars)

    """
    
    # pars = [teff,logg,metal,alpha]
    # can be 2D
    if pars.ndim==1:
        pars = np.atleast_2d(pars)
    npars = pars.shape[0]

    ngrids = len(GRIDINFO)
    
    # Loop over each star and assign it to a grid based on teff/logg
    gridindex = np.zeros(npars,int)
    newpars = pars.copy()
    for i in range(npars):
        pars1 = pars[i,:]  # teff, logg, metal, alpha
        teff1 = pars1[0]
        logg1 = pars1[1]
        metal1 = pars1[2]
        alpha1 = pars1[3]        
        if teff1 < np.min(GRIDINFO['teffrange']):
            gridindex[i] = 0
        elif teff1 > np.max(GRIDINFO['teffrange']):
            gridindex[i] = ngrids-1
        else:
            ind, = np.where((teff1 >= GRIDINFO['teffrange'][:,0]) &
                            (teff1 <= GRIDINFO['teffrange'][:,1]))
            gridindex[i] = ind[0]
                            
        # Deal with out of bounds teff/logg values
        #  set to the boundary
        teffr = GRIDINFO['teffrange'][gridindex[i],:]
        loggr = GRIDINFO['loggrange'][gridindex[i],:]
        mhr = GRIDINFO['mhrange'][gridindex[i],:]
        alphar = GRIDINFO['alpharange'][gridindex[i],:]        
        if teff1 <= (teffr[0]+1):
            newpars[i,0] = teffr[0]+5
        elif teff1 >= (teffr[1]-1):
            newpars[i,0] = teffr[1]-5
        if logg1 <= (loggr[0]+0.01):
            newpars[i,1] = loggr[0]+0.01
        elif logg1 >= (loggr[1]-0.01):
            newpars[i,1] = loggr[1]-0.01
        if metal1 <= (mhr[0]+0.002):
            newpars[i,2] = mhr[0]+0.002
        elif metal1 >= (mhr[1]-0.002):
            newpars[i,2] = mhr[1]-0.002
        if alpha1 <= (alphar[0]+0.002):
            newpars[i,3] = alphar[0]+0.002
        elif alpha1 >= (alphar[1]-0.002):
            newpars[i,3] = alphar[1]-0.002          
            
    print('Interpolating '+str(npars)+' spectra with FERRE')
            
    # Loop over gridfiles and get all of the spectra
    # that fall in its teff range
    count = 0
    flux = None
    for i in range(ngrids):
        ind, = np.where(gridindex == i)
        nind = len(ind)
        if nind==0:
            continue
        pars2 = pars[ind,:]
        newpars2 = newpars[ind,:]        

        gfile = os.path.basename(GRIDINFO['filename'][i])
        print(i+1,len(ind),gfile)
        
        # Interpolate in the FERRE grid
        fout = ferre.interp(newpars2,grid=gfile)

        # Wavelength solution
        if flux is None:
            wave = fout['wave']
            npix = len(wave)
            flux = np.zeros((npars,npix),float) 

        # Fill in the spectra
        flux[count:count+nind,:] = fout['flux']
        totflux = np.sum(flux,axis=1)
        
        count += nind

    # Put everything together
    out = {'wave':wave,'flux':flux}
    return out


def sspgrid(ages,metals,alphas,tempsave=True,outdir='./',clobber=False):
    """
    Run a grid of SSP spectra.

    Parameters
    ----------
    ages : numpy array or list
       A list or numpy array of ages in Gyr.
    metals : numpy array or list
       A list or numpy array of metallicities.
    alphas : numpy array or list
       A list or numpy array of alpha abundances [alpha/M].
    tempsave : boolean, optional
       Save the individual SSP synthetic spectra to a temporary directory.
       This allows for an easy restart if there is a crash.  Default is True.
    outdir : str, optional
       The output directory for the temporary SSP synthetic spectra.  Default
       is "./".
    clobber : bool, optional
       Overwrite existing saved SSP synthetic spectra.  Default is False.

    Returns
    -------
    wave : numpy array
       Wavelength array, 1-D.
    spectra : numpy array
       Grid of SSP synthetic spectra with size [Nages,Nmetals,Nalphas,Nwave].
    pars : numpy array
       Array of parameters with size [Nages,Nmetals,Nalphas,Nwave].

    Example
    -------

    wave,spectra,pars = sspgrid(ages,metals,alphas)

    """

    nage = len(ages)
    nmetal = len(metals)
    nalpha = len(alphas)
    nspectra = nage * nmetal * nalpha
    print('Creating {:d} SSP spectra'.format(nspectra))
    
    # age, metallicity, alpha abundance
    dt = [('age',float),('metal',float),('alpha',float),('luminosity',float)]
    pars = np.zeros((nage,nmetal,nalpha),dtype=np.dtype(dt))
    count = 0
    for i in range(nage):
        age = ages[i]
        for j in range(nmetal):
            metal = metals[j]
            for k in range(nalpha):
                alpha = alphas[k]
                print('{:d} (age,[M/H],[alpha/M])=({:.4f},{:.4f},{:.4f})'.format(count+1,age,metal,alpha))

                outfile = 'ssp_a{:.3f}m{:+.2f}a{:+.2f}.fits'.format(age,metal,alpha)
                if os.path.exists(outdir+outfile) and clobber==False:
                    print(outfile+' already exists and clobber not set')
                    spectrum,hd = fits.getdata(outdir+outfile,header=True)
                    wave = np.arange(hd['naxis1'])*hd['cdelt1']+hd['crval1']
                else:
                    try:
                        wave,spectrum = ssp(age,metal,alpha)
                    except:
                        print('CRASH!!!!')
                        print(' ')
                        import pdb; pdb.set_trace()
                        continue
                        
                # Start grid
                if count == 0:
                    nwave = len(wave)
                    spectra = np.zeros((nage,nmetal,nalpha,nwave),float)

                # Plug in the spectrum
                spectra[i,j,k,:] = spectrum
                dw = wave[1]-wave[0]
                luminosity = np.sum(spectrum*dw)
                pars['age'][i,j,k] = age
                pars['metal'][i,j,k] = metal
                pars['alpha'][i,j,k] = alpha
                pars['luminosity'][i,j,k] = luminosity

                # Write to a file in case it crashes at some point
                if os.path.exists(outdir+outfile)==False or clobber:
                    hdu = fits.PrimaryHDU(spectrum)
                    hdu.header['CRVAL1'] = wave[0]
                    hdu.header['CRPIX1'] = 1
                    hdu.header['CDELT1'] = np.round(wave[1]-wave[0],6)
                    hdu.header['DC-FLAG'] = 0
                    hdu.writeto(outdir+outfile)
                
                print(' ')
                
                count += 1

    return wave,spectra,pars

                
def ssp(age,metal,alpha,alliso=None):
    """
    Make SSP (simple stellar population)) for a given age,
    metallicity and alpha abundance.

    Parameters
    ----------
    age : float
       Age of the SSP in Gyr.
    metal : float
       Metallicity of the SSP.
    alpha : float
       The [alpha/M] abundance of the SSp.
    alliso : table, optional
       The full isochrone table.  This can save some
       time if it does not need to be imported every time.

    Returns
    -------
    wave : numpy array
       Wavelength array.
    spectrum : numpy array
       SSP synthetic spectrum array.

    Example
    -------

    wave,spectrum = ssp(1.0,-1.5,0.3)

    """

    print('Age = {:.4f} Gyr'.format(age))
    print('[M/H] = {:.2f}'.format(metal))
    print('[alpha/Fe] = {:.2f}'.format(alpha))    

    # Salaris correction
    metal_salaris = metal + np.log10(0.659*(10**alpha)+0.341)
    print('[M/H]_Salaris = {:.2f}'.format(metal_salaris))    
    
    # --- Isochrones ---
    if alliso is None:
        alliso = Table.read(utils.datadir()+'ssp_isochrones.fits.gz')
        for c in alliso.colnames: alliso[c].name = c.upper()

    umetal = np.unique(alliso['METAL'])
    uage = np.unique(alliso['AGE']/1e9)
    bestmetal,bestmetalind = dln.closest(umetal,metal_salaris)
    bestage,bestageind = dln.closest(uage,age)

    print('Closest isochrone values')
    print('Age = {:.4f} Gyr'.format(bestage))
    print('[M/H] = {:.2f}'.format(bestmetal))
    print('[alpha/Fe] = {:.2f}'.format(alpha))    
    
    # Get the isochrone we want
    isoind, = np.where((alliso['METAL']==bestmetal) & (alliso['AGE']/1e9==bestage))
    iso = alliso[isoind]
    
    # --- Create synthetic photometry ---
    stab = synth(iso,[],minlabel=0,maxlabel=9,nstars=100000)
    totmass = np.sum(stab['MINI'])
    
    # Make 2-D bins so we don't have to interpolate so many spectra
    data = stab['LOGTE'],stab['LOGG']
    # Teff : 2000 to 60,000 K
    # logg : -2.0 to 6.0
    bins = np.arange(3.30,4.78,0.01),np.arange(-2.0,6.0,0.10)    
    binout = binned_statistic_2d(stab['LOGTE'],stab['LOGG'],stab['LOGG'],
                                 statistic='count',bins=bins,expand_binnumbers=True)
    result, teff_edge, logg_edge, binnumber = binout
    teff_center = 10**((teff_edge[0:-1]+teff_edge[1:])*0.5)
    logg_center = (logg_edge[0:-1]+logg_edge[1:])*0.5
    binid = np.char.array(binnumber[0,:]).astype(str) + '-' + np.char.array(binnumber[1,:]).astype(str)
    bin_index = dln.create_index(binid)
    npars = len(bin_index['value'])
    totluminosity = np.zeros(npars,float)  # solar luminosities
    pars = np.zeros((npars,4),float)
    pars[:,2] = metal
    pars[:,3] = alpha
    for i,ubinid in enumerate(bin_index['value']):
        ind = bin_index['index'][bin_index['lo'][i]:bin_index['hi'][i]+1]
        nind = len(ind)
        totluminosity[i] = np.sum(10**stab['LOGL'][ind])
        ubin2d = np.array(ubinid.split('-')).astype(int)
        # values that are beyond the bounds are set to N
        if ubin2d[0]==len(teff_edge):
            ubin2d[0] -= 1
        if ubin2d[1]==len(logg_edge):
            ubin2d[1] -= 1            
        # the indices give the upper value of the "edge"
        # so we need to subtract 1 to get the correct "center"
        pars[i,0] = teff_center[ubin2d[0]-1]  
        pars[i,1] = logg_center[ubin2d[1]-1]    

        
    #from dlnpyutils import plotting as pl
    #import pdb; pdb.set_trace()

        
    # Plot the binned synthetic stars
    #from dlnpyutils import plotting as pl
    #pl.display(result.T,teff_edge,logg_edge,log=True,xflip=True,yflip=True)
    
    # Use FERRE to interpolate in the grid to the
    # synthetic photometric Teff, logg, [Fe/H], and [alpha/Fe]
    fout = ferre_interp(pars)
    wave = fout['wave']   # 1-D wavelength array
    flux = fout['flux']   # 2-D flux array [Nspec,Nwavelength]
    npix = len(wave)
    
    # Convert each spectrum to a total flux of 1 Lsun
    dw = 0.1
    lsun = 3.846e33   # erg/s
    totflux = np.sum(flux*dw,axis=1)
    scalefactor = lsun / totflux
    scaleflux = flux * scalefactor.reshape(-1,1)
    
    # Combine all of the spectra with appropriate luminosity weighting
    spectrum = np.sum(scaleflux * totluminosity.reshape(-1,1),axis=0)

    # Now normalize by the total mass
    spectrum /= totmass
    
    # Units
    # synspec returns fluxed spectra in erg/cm2/s/A
    # PARSEC luminosities are in solar luminosity units (3.846 x 10^33 erg/s)
    # each pixel is 0.1 A

    # np.sum(spectrum*dw) / lsun
    print('Final spectrum luminosity = {:.2f} Lsun of 1Msun'.format(np.sum(spectrum*dw) / lsun))
    print('Total synthetic photometry luminosity = {:.2f} Lsun'.format(np.sum(totluminosity)/totmass))
    
    return wave,spectrum
    
    
def synth(iso,bands,nstars=None,totmass=None,minlabel=1,maxlabel=8,minmass=0,maxmass=1000,
          columns=['AGE','METAL','MINI','LOGTE','LOGG','LOGL','LABEL']):
    """
    Create synthetic population with an isochrone and its IMF information.

    Parameters
    ----------
    iso : table
       The isochrone table.
    bands : list
       List of the bands to interpolate.  Can also be an empty list.
    nstars : int, optional
       Number of total stars to simulate.  Default is 1000.
    totmass : float, optional
       Total mass to use for the stellar populations.  If nstars and totmass
       are not input, then 1000 stars will be used and the total mass will be
       set accordingly.
    minlabel : int, optional
       Minimum PARSEC label to use.  Default is 1.
    maxlabel : int, optional
       Maximum PRASEC label to use.  Default is 8.
    minmass : float, optional
       Minimum stellar mass to use.  Default is 0.
    maxmass : float, optional
       Maximum stellar mass to use.  Default is 1000.
    columns : list, optional
       Columns to include in the output table.  Default is ['AGE','METAL',
       'MINI','LOGTE','LOGG','LOGL','LABEL'].

    Returns
    -------
    out : table
       Table of the simulated stars from the isochrone.

    Example
    -------

    tab = synth(iso)

    """
    
    # By default us 1000 stars
    if nstars is None and totmass is None:
        nstars = 1000

    lab = ((iso['LABEL']>=minlabel) & (iso['LABEL']<=maxlabel))
    data = iso[lab]

    # AGE and METAL columns
    if 'AGE' not in data.colnames and 'LOGAGE' in data.colnames:
        data['AGE'] = 10**data['LOGAGE']
    if 'METAL' not in data.colnames and 'MH' in data.colnames:
        data['METAL'] = data['MH']
    
    ## Mass cuts
    massind, = np.where((data['MINI'] >= minmass) & (data['MINI'] <= maxmass))
    data = data[massind]
    ndata = len(data)
    if ndata==0:
        raise ValueError('No isochrone points left after mass cut')
    
    # Total mass input, figure out the number of stars we expect
    # for our stellar mass range
    if nstars is None and totmass is not None:
        nstars = np.ceil((np.max(data['INT_IMF'])-np.min(data['INT_IMF']))*totmass)
            
    # Initialize the output catalog
    out = Table()
    out['AGE'] = np.zeros(int(nstars),float)
    for c in columns:
        if c != 'AGE':
            out[c] = 0.0
    for n in bands:  # bands to interpolate
        out[n] = 0.0        

    # PDF, probability distribution function
    # int_IMF, which is the integral of the IMF under consideration (as selected in the form, in number of stars,
    # and normalised to a total mass of 1 Msun) from 0 up to the current M_ini. Differences between 2 values of
    # int_IMF give the absolute number of stars occupying that isochrone section per unit mass of stellar
    # population initially born, as expected for the selected IMF.
    pdf = np.maximum(np.diff(data['INT_IMF'].data),1e-8)
    pdf = np.hstack((0.0, pdf))
        
    # Create normalized index array (from 0-1)
    indx = np.arange(ndata).astype(float)/(ndata-1)
    # Create cumulative distribution function
    cdf = np.cumsum(pdf)
    cdf /= np.max(cdf)

    # Get the indices in the structure from the CDF
    #interp,cdf,indx,randomu(seed,nstars),newindx
    newindx = interp1d(cdf,indx)(np.random.rand(int(nstars)))
        
    # Interpolate all of the relevant columns
    for n in out.colnames:
        if n != 'INT_IMF':
            newval = interp1d(indx,data[n])(newindx)
            out[n] = newval
    out['TEFF'] = 10**out['LOGTE']
            
    return out



import os
import sys

from setuptools import setup, find_packages, Extension
from setuptools.command.build_py import build_py as _build_py

setup(name='popstar',
      version='1.0.0',
      author = 'David Nidever',
      author_email='dnidever@montana.edu',
      url = 'https://github.com/dnidever/popstar',
      package_dir = {'': 'python'},
      packages = ["popstar"],
      install_requires=['numpy','astropy(>=4.0)','scipy','dlnpyutils(>=1.0.3)','ferre','gdown'],
      include_package_data=True
)
